#!/usr/bin/env python
from __future__ import (print_function as _pf,
                        unicode_literals as _ul,
                        absolute_import as _ai)


def printv(verbose, message):
    if verbose:
        print(message)
    else:
        pass


def calc_physics(variable, dives, depth,
                 spike_window=3, spike_method='minmax',
                 iqr=1.5,
                 savitzky_golay_window=11, savitzky_golay_order=2,
                 verbose=True, name='Physics Variable'):
    """
    This function is a standard setup for processing physics
    variables (temperature and salinity).

    The function applies a neighbourhood interquartile range (IQR)
    outlier filter, the Briggs et al. (2011) spike filter
    followed by a Savitzky-Golay smoothing function.

    For more information about the neighbourhood IQR outlier filter
    see the docs for buoyancy_glider_utils.tools.neighbourhood_iqr.

    The Savitzky-Golay filter is demonstrated well on wikipedia:
    https://en.wikipedia.org/wiki/Savitzky-Golay_filter
    """

    from pandas import Series
    from numpy import array, ma, NaN, inf
    from .cleaning import savitzky_golay, despike, outlier_bounds_iqr
    from . import optics as op

    x = array(dives)
    y = array(depth)

    # an interpolation step is added so that no nans are created.
    # Note that this interpolates on the flattened series
    if isinstance(variable, Series):
        name = variable.name

    printv(verbose, '\n' + '=' * 50 + "\n{}:".format(name))

    if iqr:
        ll, ul = outlier_bounds_iqr(variable, multiplier=iqr)
        mask = (variable < ll) | (variable > ul)
        z = variable.where(~mask)
        printv(verbose, '    Removing outliers with IQR * {}: {} obs'.format(iqr, mask.sum()))

    if spike_window:
        b, s = despike(z, spike_window, spike_method)
        z = b
        printv("    Removing spikes with rolling median (spike window={})".format(spike_window))

    if savitzky_golay_window:
        printv(verbose, "    Smoothing with Savitzky-Golay filter (window={}, order={})".format(savitzky_golay_window, savitzky_golay_order))
        z = savitzky_golay(z, savitzky_golay_window, savitzky_golay_order)

    return Series(z, name=name)


def calc_oxygen(o2raw, pressure, salinity, temperature, conversion='auto',
                spike_window=7, spike_method='median',
                savitzky_golay_window=0, savitzky_golay_order=2,
                verbose=True):
    """
    This function processes oxygen. It is assumed that either
    mL/L or umol/kg are passed as input. The units are automatically
    detected by looking at the mean ratio
    Below are some conversions to help with the Oxygen units
        µmol/l > µmol/kg * 1.025
        µmol/l > ml/l * 44.66
        µmol/l > mg/l * 31.25
    """

    from sklearn.linear_model import LinearRegression
    from numpy import isnan, abs, array, ones, c_, median
    from . cleaning import outlier_bounds_iqr, despike, savitzky_golay
    import seawater as sw
    from pandas import Series

    if isinstance(o2raw, Series):
        name = o2raw.name
    else:
        name = 'Oxygen'

    if spike_window:
        o2raw, _ = despike(o2raw, spike_window, spike_method)
        printv(verbose,
                '\n' + '=' * 50 + "\n{}:\n"
                "\tSmoothing data with despiking algorithm:\n\t"
                "    spike identification (spike window={})"
                "".format(name, spike_window))

    if savitzky_golay_window:
        printv(verbose, "\tSmoothing with Savitzky-Golay filter (window={}, order={})".format(savitzky_golay_window, savitzky_golay_order))
        o2raw = savitzky_golay(o2raw, savitzky_golay_window, savitzky_golay_order)

    o2sat = sw.satO2(salinity, temperature)
    density = sw.dens(salinity, temperature, pressure)

    if conversion == 'auto':
        # use linear regression to determine the oxygen unit
        # raw surface (<10m) O2 is regressed theoretical saturation
        # the slope of the regression will be indicative of the
        # units as theoretical saturation is always in mL/L
        # Use the min difference between the slope and known
        # conversion factors to estimate the appropriate conversion.

        # clean the data first with basic cleaning
        surf = ((pressure < 20) & ~isnan(o2raw) & ~isnan(o2sat))
        # prepare the data for linear regression
        Y = o2raw[surf].copy()
        X = c_[ones(surf.sum()), o2sat[surf]]
        # removing outliers accodring to IQR
        ll, ul = outlier_bounds_iqr(Y, multiplier=1.5)
        m = (Y > ll) & (Y < ul)
        ratios = Y[m] / X[m, 1]

        # compare the slopes
        observed_ratio = median(ratios)
        # the theoretical values have been divided by 1.025 to account for
        # the density of seawater
        theoretic_ratio = array([1, 43.5])
        ratio_diffs = abs(observed_ratio - theoretic_ratio)
        # catch if the difference is too big
        if ratio_diffs.min() > 10:
            printv(verbose,
                ('Oxygen unit could not be estimated automatically. '
                 'Do the unit conversion on the raw data before '
                 'passing it to the function. \n'
                 'Below is some info to help you\n'
                 '    µmol/l > µmol/kg * 1.025\n'
                 '    µmol/l > ml/l * 44.66\n'
                 '    µmol/l > mg/l * 31.25'))
        # otherwise do the conversion
        else:
            unit_idx = ratio_diffs.argmin()
            if unit_idx == 0:
                unit = 'mL/L'
                o2mll = array(o2raw)
            elif unit_idx == 2:
                unit = 'mg/L'
                o2mll = array(o2raw) / 31.25 * (density / 1000)
            elif unit_idx == 1:
                unit = 'umol/kg'
                o2mll = array(o2raw) / 44.66 * (density / 1000)
            else:
                printv(verbose, 'Difference is {}'.format(ratio_diffs))
            printv(verbose, '\tUnits automatically detected {}'.format(unit))
            if ratio_diffs.min() > 5:
                print('\tWARNING: Confirm units mannually as near the confidence threshold')
        o2aou = o2sat - o2mll
        o2pct = o2mll / o2sat * 100
        return o2mll, o2pct, o2aou

    else:
        print(
            "No oxygen conversion applied - user "
            "must impliment before or after running "
            "the cleaning functions.")


def calc_backscatter(bb_raw, wavelength, tempC, salt, dives, depth,
            dark_count, scale_factor, deep_method='median', iqr=3,
            profiles_ref_depth=300, deep_multiplier=1,
            spike_window=7, spike_method='median',
            return_figure=False, verbose=True):
    """
    INORMATION
    ----------
    Process the raw backscattering data from channel 2 at wavelength 700 nm.
    NOTE: This uses the same processing steps as calc_bb1.

    Use standard values for the CSIR SeaGliders to process backscatter.
        wavelength = 700nm
        theta angle of sensors = 124deg
        xfactor for theta 124 = 1.076
    This function also makes use of the flo_functions toolkit (Zhang et al. 2009)
    to calculate total backscatter.

    The following steps are applied in this sequence:
    1. find_bad_profiles  (high values below 300 m are counted as bad profiles)
    2. flo_scale_and_offset (factory scale and offset)
    3. flo_bback_total  (total backscatter based on Zhang et al. 2009)
    4. backscatter_dark_count  (based on Briggs et al. 2011)
    5. seperate_spikes  (using Briggs et al. 2011 - rolling min--max)
    6. neighbourhood_iqr  (see buoyancy_glider_utils.tools.neighbourhood_iqr)

    INPUT
    -----
    All inputs must be ungridded np.ndarray or pd.Series data
    bb_raw      raw output from backscatter 470 nm
    tempC       QC'd temperature in degC
    salt        QC'd salinity in PSU
    dives       the dive count (round is down dives, 0.5 up dives)
    depth       in metres

    wavelength                  e.g. 700 nm / 470 nm
    dark_count                  factory values from the cal sheet
    scale_factor                factory values from the cal sheet
    spike_window                the window size over which to run the despiking
    spike_method                whether to use a rolling median or combination of min+max filter
    profiles_ref_depth          the reference depth to determine bad profiles
    method                      whether to use the deep median or deep mean to determine bad profiles
    neighbour_iqr_multiplier    the multiplier for the iqr
    neighbour_depth_window      the depth window for the iqr
    neighbour_dives_window      the dive window for the iqr

    return_figure return a figure object that shows the before and after
    verbose       will print the progress of the processing

    OUTPUT
    ------
    baseline    an np.ma.masked_array with the mask denoting the filtered values
                the baseline of the backscatter as defined by Briggs et al. 2011
    spikes      the spikes of backscatter from Briggs et al. 2011
    fig         a figure object if return_figure=True else returns None
    """
    from numpy import array, nan, count_nonzero, unique
    from pandas import Series
    from . import optics as op
    from . import flo_functions as ff
    from .cleaning import despike, despiking_report, outlier_bounds_iqr

    bb_raw = Series(bb_raw.copy())
    dives = array(dives)
    depth = array(depth)
    tempC = array(tempC)
    salt = array(salt)
    name = 'bb{:.0f}'.format(wavelength)
    theta = 124  # factory set angle of optical sensors
    xfactor = 1.076  # for theta 124

    ref_depth = profiles_ref_depth
    stdev_multiplier = deep_multiplier
    method = deep_method

    dive_count = count_nonzero(unique(dives))

    printv(verbose, '\n' + '=' * 50 + "\n{}:".format(name))

    if iqr:
        ll, ul = outlier_bounds_iqr(bb_raw, multiplier=iqr)
        mask = (bb_raw < ll) | (bb_raw > ul)
        z = bb_raw.where(~mask)
        printv(verbose, '\tRemoving outliers with IQR * {}: {} obs'.format(iqr, mask.sum()))

    printv(verbose, '\tMask bad profiles based on deep values (depth={}m)'.format(ref_depth))
    bad_profiles = op.find_bad_profiles(dives, depth, bb_raw, ref_depth, stdev_multiplier, method)
    bb_raw[bad_profiles[0]] = nan

    bad_count = count_nonzero(bad_profiles[1])

    printv(verbose, '\tNumber of bad profiles = {}/{}'.format(bad_count, dive_count))
    printv(verbose, '\tZhang et al. (2009) correction')
    beta = ff.flo_scale_and_offset(bb_raw, dark_count, scale_factor)
    bbp = ff.flo_bback_total(beta, tempC, salt, theta, wavelength, xfactor)

    # This is from .Briggs et al. (2011)
    printv(verbose, '\tDark count correction')
    bbp = op.backscatter_dark_count(bbp, depth)

    printv(verbose, "\tSpike identification (spike window={})".format(spike_window))
    baseline, spikes = despike(bbp, spike_window, spike_method='median')
    baseline = Series(baseline, name="bb{:.0f}".format(wavelength))

    if not return_figure:
        return baseline, spikes
    else:
        printv(verbose, '\tGenerating figure for despiking report')
        fig = despiking_report(dives, depth, bbp, baseline, spikes, name=name)

        return baseline, spikes, fig


def calc_fluorescence(flr_raw, bbp, dives, depth, time, lat, lon, dark_count, scale_factor,
                      spike_window=7, spike_method='median',
                      night_day_group=True, sunrise_sunset_offset=1,
                      profiles_ref_depth=300, deep_multiplier=1, deep_method='median',
                      return_figure=False, verbose=True):
    """
    INFORMATION
    -----------
    This function processes Fluorescence and corrects for quenching using
    the Thomalla et al. (2017) approach.

    The standard sequence is applied:
    1. find_bad_profiles  (high Fluorescence in > 300 m water signals bad profile)
    2. fluorescence_dark_count  (factory correction)
    3. seperate_spikes  (using Briggs et al. 2011 - rolling min--max)
    5. quenching_correction  (corrects for quenching with Thomalla et al. 2017)
    6. neighbourhood_iqr  (find the outliers based on the interquartile range)

    INPUT
    -----
    All inputs must be ungridded np.ndarray or pd.Series data
    flr_raw     raw output from backscatter 470 nm
    bbp         processed backscatter from less noisy channel
    dives       the dive count (round is down dives, 0.5 up dives)
    depth       in metres
    time        as a np.datetime64 array
    lat, lon    latitude and longitude

    dark_count                  factory values from the cal sheet
    scale_factor                factory values from the cal sheet
    spike_window                the window size over which to run the despiking
    spike_method                whether to use a rolling median or combination of min+max filter
    night_day_group             If True, use night after otherwise use night before for flr:bbp ratio
    sunrise_sunset_offset       the delayed onset and recovery of quenching in hours [1] (assumes symmetrical)
    profiles_ref_depth          the reference depth to determine bad profiles
    method                      whether to use the deep median or deep mean to determine bad profiles

    return_figure               return a figure object that shows the before and after
    verbose                     will print the progress of the processing

    OUTPUT
    ------
    baseline            uncorrected, but despiked fluorescence
    quench_corrected    quench corrected fluorescence
    quench_layer        the quenching layer as a mask
    figs                figures reporting the despiking and quenching correction
    """

    from numpy import array, nan, count_nonzero, unique
    from . import optics as op
    from .cleaning import despike, despiking_report

    flr_raw = array(flr_raw)
    bbp = array(bbp)
    dives = array(dives)
    depth = array(depth)
    time = array(time)
    lat = array(lat)
    lon = array(lon)
    ref_depth = profiles_ref_depth
    stdev_multiplier = deep_multiplier
    method = deep_method

    printv(verbose, '\n' + '=' * 50 + '\nFluorescence\n\tMask bad profiles based on deep values (ref depth={}m)'.format(ref_depth))
    bad_profiles = op.find_bad_profiles(dives, depth, flr_raw, ref_depth, stdev_multiplier, method)
    flr_raw[bad_profiles[0]] = nan

    bad_count = count_nonzero(bad_profiles[1])
    dive_count = count_nonzero(unique(dives))
    printv(verbose, '\tNumber of bad profiles = {}/{}'.format(bad_count, dive_count))

    printv(verbose, '\tDark count correction')
    flr_raw -= dark_count
    flr_dark = op.fluorescence_dark_count(flr_raw, depth)
    flr_dark[flr_dark < 0] = nan

    baseline, spikes = despike(flr_dark, spike_window, spike_method='median')

    printv(verbose, '\tQuenching correction')
    quench_corrected, quench_layer = op.quenching_correction(
        baseline, bbp, dives, depth, time, lat, lon,
        sunrise_sunset_offset=1,
        night_day_group=True)

    printv(verbose, "\tSpike identification (spike window={})".format(spike_window))

    if return_figure:
        printv('\tGenerating figures for despiking and quenching report')
        figs = despiking_report(dives, depth, flr_raw, baseline.data, spikes, name='Fluorescence'),
        figs += op.quenching_report(baseline.data, quench_corrected.data, quench_layer, dives, depth),
        return baseline, quench_corrected, quench_layer, figs
    else:
        return baseline, quench_corrected, quench_layer


def calc_par(par_raw, dives, depth, time, scale_factor_wet_uEm2s, sensor_output_mV, replace_surface_depth=2, curve_max_depth=80, verbose=True):
    """
    INFORMATION
    -----------
    Calculates the theoretical PAR based on an exponential curve fit.

    The processing steps are:
    1. par_scaling  (factory cal sheet scaling)
    2. par_dark_count  (correct deep par values to 0 using 5th %)
    3. par_fill_surface  (return the theoretical curve of par based exponential fit)

    INPUT
    -----
    All inputs must be ungridded np.ndarray or pd.Series data
    par_raw     raw PAR
    dives       the dive count (round is down dives, 0.5 up dives)
    depth       in metres
    time        as a np.datetime64 array
    """

    from numpy import array
    from . import optics as op

    par_raw = array(par_raw)
    dives = array(dives)
    depth = array(depth)
    time = array(time)

    printv(verbose, '\n' + '=' * 50 + '\nPAR\n\tDark correction')

    # dark correction for par
    par_scaled = op.par_scaling(par_raw, scale_factor_wet_uEm2s, sensor_output_mV)
    par_dark = op.par_dark_count(par_scaled, dives, depth, time)
    printv(verbose, '\tFitting exponential curve to data')
    par_filled = op.par_fill_surface(par_dark, dives, depth, max_curve_depth=curve_max_depth)
    par_filled[par_filled < 0] = 0

    return par_filled


if __name__ == '__main__':
    pass
