#!/usr/bin/env python
from __future__ import (print_function as _pf,
                        unicode_literals as _ul,
                        absolute_import as _ai)


def grid_data(x, y, var, bins=None, how='mean', interp_lim=6, as_xarray=False):
    """
    This function bins the variable to set depths.
    If no depth is specified it defaults to:
        start : step : end
        ------------------------
           0  :  1.  : maxdepth
        ------------------------
    This is the sampling method typically used by the CSIR.
    This can be easily changed by specifying an array of increasing values.

    The function used to bin the data can also be set, where the default
    is the mean. This can be changed to 'median', 'std', 'count', etc...

    """
    from pandas import cut, Series
    from numpy import arange, array, c_, nanmax, unique
    from pandas import Series

    z = Series(var)
    y = array(y)
    x = array(x)

    u = unique(x).size
    s = x.size
    if (u / s) > 0.2:
        raise UserWarning(
            'The x input array must be paritally discrete (dives or dive_time). '
            '{:.0f}% of x is unique (max 20% unique allowed)'.format(u / s * 100))

    bins = arange(nanmax(y) + 1) if bins is None else bins

    labels = c_[bins[:-1], bins[1:]].mean(axis=1)
    bins = cut(y, bins, labels=labels)

    grp = Series(z).groupby([x, bins])
    grp_agg = getattr(grp, how)()
    gridded = grp_agg.unstack(level=0)
    gridded = gridded.reindex(labels.astype(float))

    if interp_lim > 0:
        gridded = gridded.interpolate(limit=interp_lim).bfill(limit=interp_lim)

    if as_xarray:
        return gridded.stack().to_xarray()
    return gridded


def time_average_per_dive(dives, time, time_units='seconds since 1970-01-01 00:00:00'):
    from pandas import Series
    from numpy import datetime64, array
    from xarray.coding.times import decode_cf_datetime

    time = array(time)
    dives = array(dives)
    if isinstance(time[0], datetime64):
        t = time.astype(float) / 1e9
    else:
        t = time

    t_num = Series(t).groupby(dives).mean()
    t_d64 = decode_cf_datetime(t_num, time_units)
    t_ser = Series(t_d64, index=t_num.index.values)
    t_ful = t_ser.reindex(index=dives).values

    return t_ful


def mask_to_depth_array(dives, depth, var):
    from numpy import r_, diff, array
    from pandas import Series
    i = r_[False, diff(var)].astype(bool)
    idx_depth = Series(array(depth)[i], index=array(dives)[i])

    return idx_depth


def merge_dimensions(df1, df2):
    """
    Merge the data from two different dimensions. This requires
    that the index of the dataframes is np.datetime64.
    """

    import numpy as np
    import pandas as pd

    def isdate(df):
        return isinstance(df.index.values[0], (np.datetime64, type(pd.NaT)))

    if isdate(df1) & isdate(df2):
        df = pd.concat([df1, df2], sort=True, join='outer').sort_index()
        df = df.interpolate(limit=3).bfill(limit=3)
        return df.loc[df1.index]
    else:
        raise UserWarning('Both dataframe indicies need to be np.datetime64')


def gaussian_kernel(*shape):
    """
    Returns a 2D array with gaussian values to be used
    in the blob_outliers detection function.
    Can be anisotripic if shape is not square
    Scaling is determined automatically

    INPUT:
        shape - can be int (isotropic), [int, int] (anisotropic)
    OUTPUT:
        array - 2D array
    """
    from matplotlib.cbook import flatten
    from numpy import exp, mgrid

    # make shape a list regardless of input
    shape = [int(a//2) for a in flatten([shape])]
    # anisotropic if len(2) else isotropic
    if len(shape) == 1:
        sx, sy = shape[0], shape[0]
    elif len(shape) == 2:
        sx, sy = shape

    # create the x and y grid
    x, y = mgrid[-sx:sx+1, -sy:sy+1]
    sigma = [sx / 8, sy / 8]  # sigma scaled by shape
    c = tuple([sx, sy])  # centre index of x and y
    g = 1 * exp(-(
        (x - x[c])**2 / (2 * sigma[0])**2 +
        (y - y[c])**2 / (2 * sigma[1])**2))
    return g


if __name__ == '__main__':

    pass
